import csv
import os
import sys
import threading
import traceback
from datetime import datetime as Datetime, timedelta as Timedelta

from flask import g

import pytz

from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

import chellow.dloads
from chellow.models import Session, Site, Snag
from chellow.utils import hh_format
from chellow.views import chellow_redirect


def content(user):
    sess = f = writer = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names("site_snags.csv", user)
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")
        writer.writerow(
            (
                "Chellow Id",
                "Site Code",
                "Site Name",
                "Snag Description",
                "Start Date",
                "Finish Date",
                "Days Since Snag Finished",
                "Duration Of Snag (Days)",
                "Is Ignored?",
            )
        )

        now = Datetime.now(pytz.utc)

        for snag in (
            sess.query(Snag)
            .join(Site)
            .filter(Snag.site != null())
            .order_by(Site.code, Snag.description, Snag.start_date, Snag.id)
            .options(joinedload(Snag.site))
        ):
            snag_start = snag.start_date
            snag_finish = snag.finish_date
            if snag_finish is None:
                duration = now - snag_start
                age_of_snag = Timedelta(0)
            else:
                duration = snag_finish - snag_start
                age_of_snag = now - snag_finish

            writer.writerow(
                (
                    str(snag.id),
                    snag.site.code,
                    snag.site.name,
                    snag.description,
                    hh_format(snag_start),
                    hh_format(snag_finish),
                    str(age_of_snag.days + age_of_snag.seconds / (3600 * 24)),
                    str(duration.days + duration.seconds / (3600 * 24)),
                    str(snag.is_ignored),
                )
            )
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    args = (g.user,)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import csv
import os
import threading
import traceback

from flask import g

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.e.computer import forecast_date
from chellow.gas.engine import GDataSource, g_contract_func
from chellow.models import GEra, GSupply, Session, Site, SiteGEra
from chellow.utils import csv_make_val, hh_format, hh_max, hh_min, req_date, req_int
from chellow.views import chellow_redirect


def content(g_supply_id, file_name, start_date, finish_date, user):
    caches = {}
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            f"g_supply_virtual_bill_{g_supply_id}.csv", user
        )
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")

        g_supply = GSupply.get_by_id(sess, g_supply_id)

        forecast_dt = forecast_date()

        prev_titles = None

        for g_era in (
            sess.query(GEra)
            .filter(
                GEra.g_supply == g_supply,
                GEra.start_date < finish_date,
                or_(GEra.finish_date == null(), GEra.finish_date > start_date),
            )
            .order_by(GEra.start_date)
        ):
            chunk_start = hh_max(g_era.start_date, start_date)
            chunk_finish = hh_min(g_era.finish_date, finish_date)
            site = (
                sess.query(Site)
                .join(SiteGEra)
                .filter(SiteGEra.g_era == g_era, SiteGEra.is_physical == true())
                .one()
            )

            ds = GDataSource(
                sess, chunk_start, chunk_finish, forecast_dt, g_era, caches, None
            )

            titles = ["MPRN", "Site Code", "Site Name", "Account", "From", "To", ""]

            output_line = [
                g_supply.mprn,
                site.code,
                site.name,
                ds.account,
                hh_format(ds.start_date),
                hh_format(ds.finish_date),
                "",
            ]

            contract_titles = g_contract_func(
                caches, g_era.g_contract, "virtual_bill_titles"
            )()
            titles.extend(contract_titles)

            g_contract_func(caches, g_era.g_contract, "virtual_bill")(ds)
            bill = ds.bill

            for title in contract_titles:
                if title in bill:
                    output_line.append(csv_make_val(bill[title]))
                    del bill[title]
                else:
                    output_line.append("")

            for k in sorted(bill.keys()):
                output_line.extend([k, bill[k]])

            if titles != prev_titles:
                prev_titles = titles
                writer.writerow([str(v) for v in titles])
            writer.writerow(output_line)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    g_supply_id = req_int("g_supply_id")
    file_name = "g_supply_virtual_bill_" + str(g_supply_id) + ".csv"
    start_date = req_date("start")
    finish_date = req_date("finish")
    args = (g_supply_id, file_name, start_date, finish_date, g.user)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from io import StringIO

import chellow.e.computer
import chellow.reports.report_87
from chellow.utils import ct_datetime, to_utc


class Sess:
    def __init__(self, *results):
        self.it = iter(results)

    def query(self, *arg):
        return self

    def join(self, *arg):
        return self

    def order_by(self, *arg):
        return self

    def filter(self, *arg):
        return self

    def scalar(self, *arg):
        return next(self.it)

    def first(self, *arg):
        return next(self.it)

    def distinct(self, *arg):
        return next(self.it)


def test_summertime(mocker):
    start_date_ct = ct_datetime(2010, 5)
    finish_date_ct = ct_datetime(2010, 5, 31, 23, 30)
    start_date = to_utc(start_date_ct)
    finish_date = to_utc(finish_date_ct)
    contract_id = 1
    user = mocker.Mock()
    user.email_address = "sfreud"

    mock_c_months_u = mocker.patch(
        "chellow.reports.report_87.c_months_u", autospec=True
    )

    MockEra = mocker.patch("chellow.reports.report_87.Era", autospec=True)
    MockEra.finish_date = finish_date
    MockEra.start_date = start_date

    mocker.patch(
        "chellow.reports.report_87.contract_func",
        autospec=True,
        return_value=lambda: ["standing_gbp"],
    )

    f = StringIO()
    sess = Sess([], [])

    chellow.reports.report_87.create_csv(f, sess, start_date, finish_date, contract_id)

    mock_c_months_u.assert_called_with(
        start_year=start_date_ct.year,
        start_month=start_date_ct.month,
        finish_year=finish_date_ct.year,
        finish_month=finish_date_ct.month,
    )



import argparse
import os.path
import signal
import sys
import time
from os import environ

from pep3143daemon import DaemonContext, PidFile

import waitress

from chellow import create_app

app = create_app()


def chellow_start(daemon):
    chellow_port = environ["CHELLOW_PORT"] if "CHELLOW_PORT" in environ else 80
    daemon.open()
    waitress.serve(app, host="0.0.0.0", port=chellow_port)


def chellow_stop(pidfile_path):
    with open(pidfile_path) as pidfile:
        pid = int(pidfile.read())
        os.kill(pid, signal.SIGTERM)
    while os.path.exists(pidfile_path):
        time.sleep(1)


def chellow_command():
    parser = argparse.ArgumentParser()
    parser.add_argument("action", choices=["start", "stop", "restart"])
    args = parser.parse_args()

    try:
        os.makedirs(app.instance_path)
    except BaseException:
        pass
    pidfile_path = os.path.join(app.instance_path, "chellow.pid")
    pidfile = PidFile(pidfile_path)
    daemon = DaemonContext(
        pidfile=pidfile, stdin=sys.stdin, stderr=sys.stderr, stdout=sys.stdout
    )

    if args.action == "start":
        chellow_start(daemon)
    elif args.action == "stop":
        chellow_stop(pidfile_path)
    elif args.action == "restart":
        chellow_stop(pidfile_path)
        chellow_start(daemon)
